"""
"""
# imports
import os
import tagit
import json
import sys

# inner-module imports
from tagit import dict_update

# exports
__all__ = ('load_settings','load_model','get_method','db_clean','db_index','db_sync','db_merge','pgm_info','db_update','db_tags')

############################ HARDCODED CONFIG ####################################

# debug mode
DEBUG=False
VERBOSE=False

# settings search location
SETTINGS_PATH = [
      '.tagitrc'                                                                                # Current directory
    , os.path.expanduser('~/.tagitrc')                                                          # Home directory
    , '/usr/share/tagit/settings'                                                               # Installation directory
    , '/usr/share/tagit/keybindings'                                                            # Installation directory (keybindings)
    , os.path.join(os.path.dirname(tagit.__file__), 'settings.json')                            # Module directory
    , os.path.join(os.path.dirname(tagit.__file__), 'keybindings.json')                         # Module directory (keybindings)
    ]

# database search location
DATABASE_PATH = [
      os.path.expanduser('~/.tagit.db')                                                         # Home directory
    , '/usr/share/tagit/database.db'                                                            # Installation directory
    , None                                                                                      # In-memory database
    ]

################################### END CONFIG ###################################


def load_settings(args):
    """Load application settings.

    The settings are loaded from the following locations, in order:

    1. As passed on command line
    2. Default settings file
    3. Default settings, hardcoded

    """
    # Load settings
    searchpaths = [args.settings] + SETTINGS_PATH
    settings = {'session': {'verbose': VERBOSE, 'debug': DEBUG}}

    for path in searchpaths[::-1]:
        if DEBUG:
            print "Loading settings from", path

        if path is not None and os.path.exists(path):
            dict_update(settings, json.load(open(path)))

    settings['session']['verbose'] = int(args.verbose or 0) + int(settings['session'].get('verbose', False)) + int(VERBOSE)
    settings['session']['debug']   = settings['session'].get('debug', False) or DEBUG
    from tagit import Settings
    return Settings(settings)

def load_model(args, settings, create=False):
    """Load a database.

    The database is loaded from the following locations, in order:

    1. As passed on command line
    2. Default database from settings
    3. Default database, hardcoded
    4. Empty/Temporary in-memory database

    If *create*, the first file-based database will be created.

    """
    # Load database
    from tagit.model import DataModel
    searchpaths = [args.database] + [os.path.expanduser(settings['session'].get('database', None))] + DATABASE_PATH
    model = None
    for path in searchpaths:
        if settings['session'].get('debug', False):
            print "Loading database from", path

        if path is not None and (create or os.path.exists(path)):
            model = DataModel(settings, path)
            break

    if model is None:
        # paths failed, load database in memory
        model = DataModel(settings, None)

    assert model is not None
    return model

def get_method(args):
    """Return the sync method from command-line *args*.
    """
    from tagit.model import DataModel

    if args.method == 'set':
        return DataModel.SYNC_SET
    elif args.method == 'merge':
        return DataModel.SYNC_MERGE
    elif args.method == 'update':
        return DataModel.SYNC_UPDATE
    else:
        raise Exception('method not understood')

def db_clean(args):
    """Index cleanup.
    """
    global settings, model # DEBUG

    # Read basics
    settings = load_settings(args)
    model = load_model(args, settings)

    try:
        model.remove_dead('/', recursive=True)
        #model.clean()
        model.save()
    except Exception, e:
        print "Error:", e

def db_index(args):
    """Index operations.
    """
    global settings, model # DEBUG

    # Read basics
    settings = load_settings(args)
    model = load_model(args, settings, create=True)

    # Handle arguments
    recursive = args.recursive
    sync_method = get_method(args)

    # Execute action
    try:
        for root in args.ROOT:
            if args.ACTION == 'sync': # Insert and Update
                model.index_dir(root, recursive, sync_method)
                model.save()

            elif args.ACTION == 'add': # Insert only
                model.index_dir(root, recursive, sync_method, insert_only=True)
                model.save()

            elif args.ACTION == 'update': # Update only
                model.index_dir(root, recursive, sync_method, update_only=True)
                model.save()

            elif args.ACTION == 'del':
                model.remove_dir(root, recursive)
                model.save()

            else:
                raise Exception('action not understood')

    except Exception, e:
        print "Error:", e

def db_sync(args):
    """File operations.
    """
    global settings, model # DEBUG

    # Read basics
    settings = load_settings(args)
    model = load_model(args, settings)

    # Handle arguments
    root = args.ROOT
    recursive = args.recursive
    sync_method = get_method(args)

    try:
        model.update_dir(root, recursive=recursive, sync_method=sync_method)
    except Exception, e:
        print "Error:", e

def db_merge(args):
    """Merge two or more databases.
    """
    global settings, model # DEBUG

    # Read basics
    settings = load_settings(args)

    # Handle arguments
    sync_method = get_method(args)
    sources = args.SOURCES
    target  = args.TARGET

    try:
        from tagit.model import DataModel

        if not os.path.exists(target):
            # create directories
            try:
                os.makedirs(os.path.dirname(target))
            except OSError: pass # path exists already

            # copy last source to target
            import shutil
            shutil.copyfile(sources.pop(), target) # reduces sources so the last item isn't handled twice

        model_trg = DataModel(settings, target)

        for path in sources[::-1]: # traverse back to front
            if not os.path.exists(path):
                raise IOError("No such file or directory: '{}'".format(path))

            # merge path with current database
            model_src = DataModel(settings, path)
            model_src.merge(model_trg, sync_method=sync_method)

        #model_trg.save()

    except Exception, e:
        print e

def pgm_info(args):
    """Print some program data."""
    if args.ACTION == 'settings':
        settings = load_settings(args)
        print json.dumps(settings, indent=4) #, sort_keys=True)

    else:
        raise Exception('action not understood')

def db_update(args):
    from tagit.tools import db_update
    settings = load_settings(args)
    verbose = settings.trace('session', 'verbose', False)
    searchpaths = [args.database] + [os.path.expanduser(settings['session'].get('database', None))] + DATABASE_PATH
    for path in searchpaths:
        if path is not None and os.path.exists(path):
            if verbose:
                print "Updating database", path

            db_update(path, args.VERSION, verbose)
            break # First pick only!


def db_tags(args):
    """

    cleanup: Remove unattached tags
    stats:   Show tag statistics
    collect: Try to merge tags

    """
    global model, settings

    # Read basics
    settings = load_settings(args)
    model = load_model(args, settings)

    try:
        if args.ACTION == 'list':
            for tag in sorted(map(unicode.title, map(unicode.lower, model.tags.get_all()))):
                print tag

        elif args.ACTION == 'collect': # Try to shorten the set of tags
            from tagit.tools import tags_collect
            tags_collect(model)
            #model.save()

        elif args.ACTION == 'hist': # Histogram
            from tagit.tools import tags_histogram
            tags_histogram(model)

        elif args.ACTION in ('lower', 'upper'): # Tags stringops
            from tagit.tools import tags_ops
            tags_ops(model, args.ACTION)
            model.save()

        elif args.ACTION == 'cleanup': # Check database integrity and optimize
            pass

        elif args.ACTION == 'stats': # Print tag statistics
            from tagit.tools import tags_stats
            tags_stats(model)
            pass

        else:
            raise Exception('Action not understood')

    except Exception, e:
        print e

## EOF ##
