"""Extended, optional image attributes.

Part of the tagit module.
A copy of the license is provided with the project.
Author: Matthias Baumgartner, 2016

"""
# STANDARD IMPORTS
import magic

# INNER-MODULE IMPORTS
from ..basics import fst, string_types
from ..token import Token_SingleValue_Parsed, register_token

# CONSTANTS
IMAGE_ATTRIBUTES = ['image', 'mime', 'orientation', 'width', 'height', 'exposure', 'focal_length', 'focal_length_35', 'aperture', 'iso', 'flash', 'rank']

# EXPORTS
__all__ = ('Attributes_SQLite', 'IMAGE_ATTRIBUTES')

## CODE ##

class Attributes(object):
    pass

class Attributes_SQLite(Attributes):
    def __init__(self, meta_adapter, conn):
        self.conn = conn
        self.meta_adapter = meta_adapter

    def set_from_path(self, db_id, path):
        """Get attributes from EXIF and write them into the database."""
        metrics = self.meta_adapter.get_metrics(path)
        width, height, orientation = self.meta_adapter.get_dimensions(path)
        metrics.update({'width': width, 'height': height, 'orientation': orientation})
        metrics.update({'mime': magic.from_file(path, mime=True)})
        if 'exposure' in metrics: metrics['exposure'] = 1.0 / metrics['exposure']

        query = "INSERT OR REPLACE INTO attribute ("
        query += 'image, '
        query += ', '.join(metrics.keys())
        query += ') VALUES ('
        query += '?, '
        query += ', '.join('?' * len(metrics))
        query += ')'

        self.conn.execute(query, [db_id] + metrics.values())

    def set(self, image, attribute, value):
        if attribute not in IMAGE_ATTRIBUTES: raise Exception('Invalid attribute')
        query = "UPDATE attribute SET {0} = ? WHERE attribute.image = (SELECT id FROM image WHERE path = ?)".format(attribute) # FIXME: Dangerous SQL!
        self.conn.execute(query, (value, image))
        return True

    def get(self, image, attribute):
        if attribute not in IMAGE_ATTRIBUTES: raise Exception('Invalid attribute')
        query = "SELECT {0} FROM attribute JOIN image on attribute.image = image.id WHERE image.path = ?".format(attribute) # FIXME: Dangerous SQL!
        metrics = self.conn.execute(query, (image, )).fetchall()
        if len(metrics) > 0:
            value = fst(metrics[0])
            if value is None or value == 0:
                return None
            return value

        return None

    # Getters
    # FIXME: There's surely an easier way via DB abstraction...
    def orientation(self, image):       return self.get(image, 'orientation')
    def mime(self, image):              return self.get(image, 'mime')
    def width(self, image):             return self.get(image, 'width')
    def height(self, image):            return self.get(image, 'height')
    def exposure(self, image):          return self.get(image, 'exposure')
    def focal_length(self, image):      return self.get(image, 'focal_length')
    def focal_length_35(self, image):   return self.get(image, 'focal_length_35')
    def aperture(self, image):          return self.get(image, 'aperture')
    def iso(self, image):               return self.get(image, 'iso')
    def flash(self, image):             return self.get(image, 'flash')
    def rank(self, image):              return self.get(image, 'rank')

    # Setters
    # FIXME: There's surely an easier way via DB abstraction...
    def set_orientation(self, image, value):       self.set(image, 'orientation', value)
    def set_mime(self, image, value):              self.set(image, 'mime', value)
    def set_width(self, image, value):             self.set(image, 'width', value)
    def set_height(self, image, value):            self.set(image, 'height', value)
    def set_exposure(self, image, value):          self.set(image, 'exposure', value)
    def set_focal_length(self, image, value):      self.set(image, 'focal_length', value)
    def set_focal_length_35(self, image, value):   self.set(image, 'focal_length_35', value)
    def set_aperture(self, image, value):          self.set(image, 'aperture', value)
    def set_iso(self, image, value):               self.set(image, 'iso', value)
    def set_flash(self, image, value):             self.set(image, 'flash', value)
    def set_rank(self, image, value):              self.set(image, 'rank', value)


    def query(self, attributes=None, values=None):
        """
        """
        joins, conds, binds = [], [], []
        if attributes is None or values is None or len(attributes) == 0:
            return joins, conds, binds

        assert len(attributes) == len(values)

        joins.append('attribute ON image.id = attribute.image')
        for attr, val in zip(attributes, values):
            if attr not in IMAGE_ATTRIBUTES: raise Exception('Invalid attribute')
            is_list = False
            try:
                len(val)
                is_list = not isinstance(val, string_types)
            except TypeError:
                is_list = False

            if is_list and len(val) == 2: # Range
                lo, hi = val
                conds.append('attribute.{0} >= ?'.format(attr))
                conds.append('attribute.{0} <= ?'.format(attr))
                binds.append(lo)
                binds.append(hi)
                # array
            elif not is_list:
                # single value
                conds.append('attribute.{0} = ?'.format(attr))
                binds.append(val)

        return joins, conds, binds


class Token_Attribute(Token_SingleValue_Parsed):
    attribute = None
    def display(self):
        return self.attribute
    def get(self):
        return (self.attribute, self._value)
    @staticmethod
    def match(obj):
        return isinstance(obj, Token_Attribute)

class Token_Flash(Token_Attribute):
    attribute = 'flash'
    def __init__(self):
        super(Token_Flash, self).__init__(False)
    @staticmethod
    def from_string(string): return Token_Flash().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Flash)

class Token_Attribute_Range(Token_Attribute):
    def __init__(self):
        super(Token_Attribute_Range, self).__init__([0, 0])
    @staticmethod
    def match(obj):
        return isinstance(obj, Token_Attribute_Range)

class Token_Rank(Token_Attribute_Range):
    attribute = 'rank'
    @staticmethod
    def from_string(string): return Token_Rank().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Rank)

class Token_Height(Token_Attribute_Range):
    attribute = 'height'
    @staticmethod
    def from_string(string): return Token_Height().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Height)

class Token_Width(Token_Attribute_Range):
    attribute = 'width'
    @staticmethod
    def from_string(string): return Token_Width().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Width)

class Token_Mime(Token_Attribute_Range):
    attribute = 'mime'
    @staticmethod
    def from_string(string): return Token_Mime().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Mime)

class Token_Exposure(Token_Attribute_Range):
    attribute = 'exposure'
    @staticmethod
    def from_string(string): return Token_Exposure().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Exposure)

class Token_Focal_Length_35(Token_Attribute_Range):
    attribute = 'focal_length_35'
    @staticmethod
    def from_string(string): return Token_Focal_Length_35().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Focal_Length_35)

class Token_Focal_Length(Token_Attribute_Range):
    attribute = 'focal_length'
    @staticmethod
    def from_string(string): return Token_Focal_Length().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Focal_Length)

class Token_Aperture(Token_Attribute_Range):
    attribute = 'aperture'
    @staticmethod
    def from_string(string): return Token_Aperture().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_Aperture)

class Token_ISO(Token_Attribute_Range):
    attribute = 'iso'
    @staticmethod
    def from_string(string): return Token_ISO().update(string)
    @staticmethod
    def match(obj): return isinstance(obj, Token_ISO)

register_token('attribute', Token_Attribute)
register_token('height', Token_Height)
register_token('width', Token_Width)
register_token('mime', Token_Mime)
register_token('exposure', Token_Exposure)
register_token('focal_length_35', Token_Focal_Length_35)
register_token('focal_length', Token_Focal_Length)
register_token('aperture', Token_Aperture)
register_token('iso', Token_ISO)
register_token('flash', Token_Flash)
register_token('rank', Token_Rank)

## EOF ##
