"""

.. TODO::
    What happens if I query 10k entries in the database?
        * How does this affect the query?
        * How does this affect the result?
    What happens if I run operations on many (10k) images?
        * Index a directory (or directories) with many images
        * Update a directory (or directories) with many images
    Several functions need to be implemented
    Some code is functional but not nice to read or bad style and should be reviewed
    Exception handling
    Logging
    Check settings (update-on-load, update-on-read, ...)

Terminology
-----------

Targets:

* file      : A single file
* files     : A list of files
* dir       : A single directory
* path      : An arbitrary path. Need not to exist nor is it defined if it's a dir or a file

Actions:

* update    : Writes to the files
* index     : Writes to the database

Modi:

* sync      : Merge information (i.e. create the union)
* update    : Pick the newer out of two information sets.
* set       : Overwrite one set of information with another (the order is defined)


"""

# STANDARD IMPORTS
import os.path
import os
import sqlite3
import time
import itertools
import sys
import re
import magic

# INNER-MODULE IMPORTS
from ..basics import *
from db_path import DBpath
from thumbnail import Thumbnailer_Chain
from tags import Tags_SQLite
from ..thread_pool import ThreadPool
from extractor import Extractor_Aggregator_all
from attributes import Attributes_SQLite
from ..token import token_factory

Token_Tag = token_factory('tag')
Token_Include = token_factory('include')
Token_Exclude = token_factory('exclude')
Token_Attribute = token_factory('attribute')
Token_Path = token_factory('path')


check_version(sqlite3.version_info, (2, 6, 0))

# CONSTANTS
DB_VERSION = 5
DB_SCHEMA = os.path.join(os.path.dirname(__file__), 'schema', 'scheme-{0:04d}.sql'.format(DB_VERSION))
SUPPORTED_MIME_TYPES = ('image/jpeg')
NUM_THUMBNAIL_WORKERS = 5
WORK_DIR_CHECKPOINT = 20

# EXPORTS
__all__ = ('DataModel', 'DB_VERSION', 'SUPPORTED_MIME_TYPES')

## CODE ##

class DataModel(object):
    """Open a database from the file *path* or memory (if None, the default, or empty).
    If a path is given but does not exist, a new database will be created.
    Version constraints on existing databases may apply.
    """

    # STATICS
    SYNC_UNION  = 'union'           # Pick union of two sets
    SYNC_MERGE  = SYNC_UNION        # Pick union of two sets
    SYNC_SET    = 'set'             # Pick foreign set
    SYNC_UPDATE = 'update'          # Pick newer set

    def __init__(self, settings, path=None, meta_adapter=None):
        # check args
        if meta_adapter is None:
            from meta_adapter import MetaAdapter_PyExiv2 # NOTE: Currently there's only one
            meta_adapter = MetaAdapter_PyExiv2()


        # set attributes
        self.query_sort = "image.t_image_create ASC" # FIXME: Not really a feature yet (not configurable)
        self.meta_adapter = meta_adapter
        self.settings = settings
        self.console = Console(settings.trace('session', 'verbose', False)) # Some debug code in here, can you spot it?
        self.thumb_workers = ThreadPool(settings.trace('model', 'thumbnail', 'workers', NUM_THUMBNAIL_WORKERS))
        self.extractor = Extractor_Aggregator_all()

        # load the database
        self.load(path)

    def _load_thumbnailer(self):
        """Set up the thumbnail load/store mechanism, according to the config.
        """

        thumb_chain = []
        if self.settings.trace('model', 'thumbnail', 'use_file', False):
            from thumbnail import Thumbnailer_FS
            thumb_chain.append(Thumbnailer_FS(os.path.expanduser(self.settings.trace('model', 'thumbnail', 'path', '/tmp'))))

        if self.settings.trace('model', 'thumbnail', 'use_image', False):
            from thumbnail import Thumbnailer_Exif
            thumb_chain.append(Thumbnailer_Exif(self.meta_adapter))

        if self.settings.trace('model', 'thumbnail', 'use_database', False):
            from thumbnail import Thumbnailer_Database
            thumb_chain.append(Thumbnailer_Database(self.conn))

        return Thumbnailer_Chain(*thumb_chain)

    ###########################################################################
    ####                        CONNECTION HANDLING                        ####
    ###########################################################################

    def load_database(self, path=None):
        """Load a database from *path*.
        If *path* is None, the database is created in the memory.
        If *path* does not exist, a new database is created.
        """
        # args
        load_schema = path is None or not os.path.exists(path)
        path = path or ':memory:'

        # connect
        conn = sqlite3.connect(path, check_same_thread=not self.settings.trace('session', 'debug', False))
        # If the UI runs in a different thread from the model creation / database
        # connection, SQLite will complain. The parameter *check_same_thread=False*
        # can in this case be passed to *sqlite3.connect* to disable the check.
        # Errors might occur in this case.

        # load and verify schema
        if load_schema:
            from hashlib import sha1
            from time import time
            # schema
            schema = ' '.join(open(DB_SCHEMA, 'r').readlines())
            conn.executescript(schema) # Commits automatically, i.e. writes the file with the correct schema.
            # metadata
            instid = sha1(path + str(time())).hexdigest()
            conn.execute('INSERT INTO metadata (id, name, t_created) VALUES (?, ?, ?)', (instid, path, time()))
            # write changes
            conn.commit() # For explicity and has_changes

        else:
            version = fst(conn.execute('PRAGMA user_version').fetchone())
            if version != DB_VERSION:
                raise Exception('Database version mismatch. Expected {1}, found {0}. Run "tagger update {1}" to upgrade the database.'.format(version, DB_VERSION))

        ## SQLite config

        # regex function
        def regexp(pattern, literal):
            """Simple case-insensitive pattern matching."""
            return re.match(pattern, literal, re.I) is not None
        conn.create_function('regexp', 2, regexp) # Enable REGEXP operator

        # pragmas
        conn.execute("PRAGMA case_sensitive_like = false") # Make LIKE case-insensitive
        conn.execute("PRAGMA foreign_keys = 1") # Enable foreign keys
        conn.execute("PRAGMA journal_mode = TRUNCATE") # Set to truncate (faster) instead of delete (default)
        # FIXME: Might want to check if foreign_keys are (still) disabled / not available.

        # return
        return conn

    def save(self, path=None):
        """Save changes to the database.

        If *path* is set, creates a new database with the current content
        at that location. The new database is returned, the old one is neither
        changed (i.e. reverted) nor saved.

        If *path* is not set, changes to the database are commited.

        If 'write-through' is True the changes are saved to the file as well
        (-> update_files(sync_method=SYNC_SET))

        """
        if path is not None:
            # save the database to another file location.
            if os.path.exists(path):
                raise Exception('File exists. If you want to merge two existing databases, use the corresponding command-line tool')

            fork = self.merge(DataModel(path), self.SYNC_SET)
            fork.commit()
            return fork

        # Save the changes to the current database
        self.conn.commit()
        self.commited_changes = self.conn.total_changes

        # remove files in candidates list (removal_candidates)
        for image in self.removal_candidates:
            self.thumb.remove(image)

        # TODO: File changes (write-through)
        return self

    def has_changes(self):
        """Return True if the database has unsaved changes.
        """
        return self.conn.total_changes > self.commited_changes

    def has_file_connection(self):
        """Return True if the database is stored in a file.
        As opposed to in-memory database.
        """
        return self.conn_path is not None and self.conn_path != ':memory:'

    def load(self, path=None):
        """Open a database and re-initialize the instance.
        """
        # load the database
        self.commited_changes = 0
        self.conn_path = path
        self.conn = self.load_database(path) # TODO: Catch exception

        # set db-dependent attributes
        self.db_path = DBpath(self.settings, self.conn)
        self.thumb = self._load_thumbnailer()
        self.tags = Tags_SQLite(self.conn)
        self.attributes = Attributes_SQLite(self.meta_adapter, self.conn)
        self.removal_candidates = []

    def rollback(self):
        """Undo changes since the last save.
        """
        self.conn.rollback()
        self.commited_changes = self.conn.total_changes
        self.removal_candidates = []

    def get_version(self):
        """Return the database version."""
        return fst(self.conn.execute('PRAGMA user_version').fetchone())


    ###########################################################################
    ####                        DATABASE OPERATIONS                        ####
    ###########################################################################

    def merge(self, other, sync_method=SYNC_MERGE):
        """Merge this database into *other*.
        The *other* database will be changed, not this instance.
        Changes are not commited.
        Indexed images are merged.
        If an image is indexed in both, the basic information is taken from the later one.
        Tags are always merged according to sync_method.
        SYNC_SET then means to use the information of this database.
        """
        for img, db_access, db_update in self.conn.execute('SELECT path, t_db_access, t_db_update FROM image'):
            if not other._contains(img):
                # add image
                other.conn.execute("INSERT INTO image (path, t_db_access, t_db_update) VALUES (?, ?, ?)", (path, db_access, db_update))
                other.tags.add([img], self.tags.get(img))

            else: # Merge images
                # update basic information
                other_access, other_update = other.conn.execute('SELECT t_db_access, t_db_update FROM image WHERE path = ?', (img, )).fetchone()

                # merge tags
                if sync_method == self.SYNC_SET:
                    other.tags.set([img], self.tags.get(img))
                elif sync_method == self.SYNC_UNION:
                    other.tags.add([img], self.tags.get(img))
                elif sync_method == self.SYNC_UPDATE:
                    if t_db_update > other_update:
                        # same as SYNC_SET
                        other.tags.set([img], self.tags.get(img))
                else:
                    raise Exception('sync_method not understood')

            # Update time information
            db_time = time.time()
            other.conn.execute('UPDATE image SET t_db_access = ?, t_db_update = ? WHERE path = ?', (db_time, db_time, img))

            # FIXME: Merge thumbs
            # FIXME: Merge attributes

        other.save()
        return other


    ###########################################################################
    ####                     FILE/DATABASE OPERATIONS                      ####
    ###########################################################################

    ############### INTERNALS ###############

    def __work_files(self, work_file, files, common_root=None):
        """Index or reindex may *files*.
        The database will not be saved.
        """
        if common_root is not None:
            files = map(lambda fn: os.path.join(common_root, fn), files)

        num_processed, num_skipped = 0, 0
        for path in files:
            if os.path.exists(path) and os.path.isfile(path):
                try:
                    work_file(path)
                    num_processed += 1
                except Exception, e:
                    num_skipped += 1
                    if self.settings.trace('session', 'debug', False):
                        print "Error", e
            else:
                num_skipped += 1

        return num_processed, num_skipped

    def __work_dir(self, work_file, path, recursive=False):
        """Index or reindex files in *path*.
        The database will not be saved!
        """
        num_processed, num_skipped = 0, 0
        path = os.path.realpath(path)

        if not os.path.exists(path):
            # TODO: Exception management
            raise Exception('Path does not exist')

        if os.path.isfile(path):
            # Special case: *path* is actually a file
            try:
                work_file(path)
                num_processed += 1
            except Exception, e:
                num_skipped += 1
                if self.settings.trace('session', 'debug', False):
                    print "Error", e

        elif os.path.isdir(path):
            if recursive:
                # walk through files and subdirs of *dir*.
                for root, dirs, files in os.walk(path, followlinks=self.settings.trace('model', 'follow-symlinks', False)):
                    # TODO: Remove blacklisted directories from *dirs*
                    for filename in files:
                        try:
                            work_file(os.path.join(root, filename))
                            num_processed += 1
                            if WORK_DIR_CHECKPOINT > 0 and num_processed % WORK_DIR_CHECKPOINT == 0:
                                if self.settings.trace('session', 'verbose', False):
                                    print "------------------------------------- SAVING ----------------------------------"
                                self.save() # FIXME: Wouldn't be necessary for update
                        except Exception, e:
                            num_skipped += 1
                            if self.settings.trace('session', 'debug', False):
                                print "Error", e
            else:
                # process whole directory
                num_processed, num_skipped = self.__work_files(work_file, os.listdir(path), common_root=path)

        else:
            raise Exception('Can only handle files and dirs')

        return num_processed, num_skipped


    ############### INDEXING ###############

    def index_file(self, path, sync_method=SYNC_SET, insert_only=False, update_only=False):
        """Index or reindex *path*.
        The database will not be saved!
        """
        if insert_only and update_only: # Mutually exclusive!
            raise Exception('insert_only and update_only are mutually exclusive')

        # Check path properties
        path = os.path.realpath(path)
        if not os.path.exists(path) or not os.path.isfile(path):
            raise Exception('Can only work on files')


        self.console.title(truncate_dir(path, 4), 'Indexing ')

        # Check if image file
        if magic.from_file(path, mime=True) not in SUPPORTED_MIME_TYPES:
            self.console.fail()
            raise Exception('File type not supported')

        # Check *path* in database
        db_id, db_path, db_update = self.conn.execute("SELECT id, path, t_db_update FROM image WHERE path = ?", (path, )).fetchone() or (None, None, None)
        if db_id is not None:
            if insert_only: # Abort if update not requested
                self.console.ignored()
                return

            tags = []
            # update; check sync_method
            if sync_method == self.SYNC_SET:
                # pick image data
                tags = self.meta_adapter.get(path)
            elif sync_method == self.SYNC_UNION:
                # merge two sets
                tags_file = self.meta_adapter.get(path)
                tags_db   = self.tags.get(path)
                tags = union(tags_file, tags_db)
                # NOTE: Optimize: Only add tags!
            elif sync_method == self.SYNC_UPDATE:
                # pick image data if t_file > t_database
                if os.stat(path).st_mtime > db_update:
                    tags = self.meta_adapter.get(path)
                else:
                    tags = self.tags.get()
            else:
                raise Exception('sync_method not understood')

            # Automatically generated tags
            tags = tags + self.extractor.tags(path, self.settings, self.meta_adapter) # For some reason, writing tags += self.extractor... causes a Segfault.

            # Write tags
            self.tags.set([path], tags)

            # Update time information
            db_time = time.time()
            self.conn.execute('UPDATE image SET t_db_access = ?, t_db_update = ? WHERE id = ?', (db_time, db_time, db_id))

            # read attributes
            self.attributes.set_from_path(db_id, path)

        else:
            if update_only: # Abort if insert not requested
                self.console.ignored()
                return

            # insert
            # basic image entry
            db_time = int(time.time())
            img_time = time.mktime(self.meta_adapter.get_date(path).timetuple())
            db_id = self.conn.execute("INSERT INTO image (path, t_db_access, t_db_update, t_image_create) VALUES (?, ?, ?, ?)", (path, db_time, db_time, img_time)).lastrowid

            # disregard sync_method, pick image data in any case.
            tags = self.meta_adapter.get(path)
            tags = tags + self.extractor.tags(path, self.settings, self.meta_adapter) # Automatically generated tags
            self.tags.add([path], tags)

            # read attributes
            self.attributes.set_from_path(db_id, path)

        # Create thumbnail
        if self.settings.trace('model', 'thumbnail', 'on_index', False):
            resolutions = self.settings.trace('model', 'thumbnail', 'size', (400, ))
            def task():
                # FIXME: check self.thumb.has(path) -> any not all!
                # But we don't want to re-create thumbs (especially exif!)
                for res in resolutions:
                    self.thumb.set(path, self.thumb.create(path, (res, res)))

            self.thumb_workers.add_task(task)

        self.console.ok()

    def index_files(self, files, sync_method=SYNC_SET, insert_only=False, update_only=False, common_root=None):
        """Index or reindex may *files*.
        The database will not be saved.
        """
        work_file = lambda path: self.index_file(path, sync_method=sync_method, insert_only=insert_only, update_only=update_only)
        return self.__work_files(work_file, files, common_root=common_root)

    def index_dir(self, path, recursive=False, sync_method=SYNC_SET, insert_only=False, update_only=False):
        """Index or reindex files in *path*.
        The database will not be saved!
        """
        if self.settings.trace('session', 'verbose', False):
            t_start = time.time()

        work_file = lambda path: self.index_file(path, sync_method=sync_method, insert_only=insert_only, update_only=update_only)
        num_indexed, num_skipped = self.__work_dir(work_file, path, recursive)

        if self.settings.trace('session', 'verbose', False):
            t_end = time.time()
            print ""
            print "   Indexed {0} files.".format(num_indexed)
            print "   Skipped {0} files.".format(num_skipped)
            print "   {0} database writes".format(self.conn.total_changes - self.commited_changes)
            print "   Elapsed time: {0} seconds".format(round(t_end - t_start, 2))
            print ""



    ############### UPDATING ###############

    def update_file(self, path, sync_method=SYNC_SET):
        """Update all files in *path* with information from the database.
        Files will be changed.
        """
        if not os.path.exists(path) or not os.path.isfile(path):
            raise Exception('Can only work on files')

        self.console.title(truncate_dir(path, 4), 'Updating ')

        # Check *path* in database
        db_image = self.conn.execute("SELECT id, path, t_db_update FROM image WHERE path = ?", (path, )).fetchone()
        if db_image is None:
            self.console.fail()
            raise Exception('Image not in database')

        tags = None
        if sync_method == self.SYNC_SET:
            # Pick database data
            tags = self.tags.get(path)
        elif sync_method == self.SYNC_UNION:
            # merge two sets
            tags_db = self.tags.get(path)
            tags_file = self.meta_adapter.get(path)
            tags = union(tags_db, tags_file)
        elif sync_method == self.SYNC_UPDATE:
            # pick database data if t_database > t_file
            if db_image[2] > os.stat(path).st_mtime:
                tags = self.tags.get(path)
        else:
            self.console.fail()
            raise Exception('sync_method not understood')

        if tags is not None:
            # Write tags to IPTC
            self.console.ok()
            self.meta_adapter.set(path, tags)
        else:
            self.console.ignored()

    def update_files(self, files, sync_method=SYNC_SET, common_root=None):
        """Update all files in *path* with information from the database.
        Files will be changed.
        """
        work_file = lambda path: self.update_file(path, sync_method=sync_method)
        return self.__work_files(work_file, files, common_root=common_root)

    def update_dir(self, path, recursive=False, sync_method=SYNC_SET):
        """Update all files in *path* with information from the database.
        If *path* is a directory, all files in that directory are updated.
        If *path* is a directory and *recursive* is true, all subdirectories will
        be changed as well.
        Files will be changed.
        """
        if self.settings.trace('session', 'verbose', False):
            t_start = time.time()

        work_file = lambda path: self.update_file(path, sync_method=sync_method)
        self.__work_dir(work_file, path, recursive)

        if self.settings.trace('session', 'verbose', False):
            t_end = time.time()
            print ""
            print "   Updated {0} files.".format(num_indexed)
            print "   Skipped {0} files.".format(num_skipped)
            print "   Elapsed time: {0} seconds".format(round(t_end - t_start, 2))
            print ""



    ############### SYNCING ###############

    def sync_file(self, path):
        """Synchronize file *path* with the database.
        Creates an UNION of file and database and writes to both.
        Files will be changed, database not saved!
        """
        self.index_file(path, self.SYNC_UNION, update_only=False, insert_only=False)
        self.update_file(path, self.SYNC_UNION)

    def sync_files(self, files):
        """Synchronize all files in *files* with the database.
        Creates an UNION of file and database and writes to both.
        Files will be changed, database not saved!
        """
        self.index_files(files, self.SYNC_UNION, insert_only=False, update_only=False) # FIXME: OK to proceed on failed files?
        self.update_files(files, self.SYNC_UNION)

    def sync_dir(self, path, recursive):
        """Synchronize all files in *path* with the database.
        Creates an UNION of file and database and writes to both.
        Files will be changed, database not saved!
        """
        # Essentially, it's this:
        self.index_dir(path, recursive, self.SYNC_UNION)
        self.update_dir(path, recursive, self.SYNC_UNION)


    ############### REMOVING ###############

    def remove_image(self, image):
        """Remove *image* from the database.
        """
        # Remove tags
        self.tags.remove_all([image])

        # Remove attributes
        #self.attributes.delete([image]) # Taken care of by foreign keys

        # Remove image
        self.conn.execute("DELETE FROM image WHERE path = ?", (image, ))

        # remove thumbnail
        self.removal_candidates.append(image)
        # Remove the thumbnail only after next save
        # Otherwise, if not saved, thumbs are removed but imgs still exist.

    def remove_images(self, images):
        """Remove *images* from the database.
        """
        for img in images:
            self.remove_image(img)

    def remove_dir(self, path, recursive=False):
        """Remove images in *path* from the database.
        Traverses subdirectories if *recursive* is set.
        *path* has to be a directory.
        """
        if recursive:
            path = path + '%' # Allow path to continue after prefix
            # FIXME: Since paths are not properly modeled, entries like
            # '../../../' might occur, so files outside *path* might be deleted.
            images = self.conn.execute("SELECT path FROM image WHERE path LIKE ?", (path, ))
            images = images.fetchall()
            images = map(fst, images)

        else:
            # remove all images in *path*
            images = self.db_path.listfiles(path)

        self.remove_images(images)

    def remove_dead(self, path, recursive=False):
        """Remove inexistent files from the database.
        """
        to_be_removed = []
        if recursive:
            for root, dirs, files in self.db_path.walk(path):
                to_be_removed += filter(lambda p: not os.path.exists(os.path.join(root, p)), files)

        else:
            files = self.db_path.listfiles(path)
            to_be_removed = filter(lambda p: not os.path.exists(p), files)

        self.conn.execute("DELETE FROM image WHERE path = ?", to_be_removed)


    ###########################################################################
    ####                        DATABASE QUERIES                           ####
    ###########################################################################

    def query(self, tokens=None):
        """Return all images which satisfy all constraining *tokens*.
        Tokens is a list of tags (exact match).
        """
        if tokens is None or len(tokens) == 0:
            results = self.conn.execute("SELECT image.path FROM image ORDER BY {}".format(self.query_sort))

        else:
            # Ordinary tags
            tags = map(Token_Tag.get, filter(Token_Tag.match, tokens))
            joins, conds, binds = self.tags.query(tags)
            #joins, conds, binds = map(operator.add, zip((joins, conds, binds), self.tags.query(tags))) # Hehehe

            # Attributes
            attrs = map(Token_Attribute.get, filter(Token_Attribute.match, tokens))
            j, c, b = self.attributes.query(*zip(*attrs))
            joins, conds, binds = joins + j, conds + c, binds + b

            # Include / Exclude
            for tok in tokens:
                if Token_Exclude.match(tok) and len(tok.get()) > 0:
                    conds.append('image.path NOT IN (' + ','.join('?' * len(tok.get())) + ')')
                    binds.extend(tok.get())
                elif Token_Include.match(tok) and len(tok.get()) > 0:
                    conds.append('image.path IN (' + ','.join('?' * len(tok.get())) + ')')
                    binds.extend(tok.get())
                elif Token_Path.match(tok) and len(tok.get()) > 0:
                    conds.append('image.path LIKE ?')
                    binds.append(tok.get())

            # Build query from parts
            query = "SELECT image.path"
            query += ' FROM image '
            if len(joins) > 0:
                query += 'JOIN ' + ' JOIN '.join(joins)
            if len(conds) > 0:
                query += ' WHERE ' + ' AND '.join(conds)
            query += " ORDER BY {}".format(self.query_sort)

            # Run query
            results = self.conn.execute(query, binds)

        return map(fst, results.fetchall())

    def get_metadata(self, image):
        """Return all data of *image*.
        """
        return self.tags.get(image)

    def get_image(self, image, resolution=None):
        """Return an openable image from *image*.
        """
        # FIXME: Can create a temporary file from object-based thumbs.
        # But then, have take care of removing temporary files.
        # Hint: tempfile.NamedTemporaryFile

        # Return available file-based thumbnail.
        if self.thumb.has(image):
            thumb = self.thumb.get(image, resolution)
            # FIXME: If object-based thumb available only (exif), copy to file-based source and return that one (currently only done if 'on_load' is set).
            # FIXME: has is ignorant of size - might want to create a larger thumb!
            if isinstance(thumb, string_types):
                return thumb

        # Create thumbnail; Return file-based thumbnail only.
        if self.settings.trace('model', 'thumbnail', 'on_load', False) and os.path.exists(image):
            selected = None
            for res in self.settings.trace('model', 'thumbnail', 'size', (400, )):
                thumb = self.thumb.set(image, self.thumb.create(image))
                if isinstance(thumb, string_types):
                    selected = thumb

            if selected is not None:
                return thumb

        # Return original
        return image

    def num_images(self):
        """Return the total number of images.
        """
        return fst(fst(self.conn.execute("SELECT count(*) FROM image").fetchall()))

    def get_creation_date(self, image):
        """Return the creation date of an image.
        """
        res = self.conn.execute("SELECT t_image_create FROM image WHERE path = ?", (image, )).fetchall()
        if len(res) > 0:
            return fst(fst(res))
        else:
            return 0

    def get_instance_infos(self):
        """Return sender id and name as tuple."""
        return self.conn.execute('SELECT id, name FROM metadata').fetchone()


    ###########################################################################
    ####                           INTERNALS                               ####
    ###########################################################################

    def _contains(self, image):
        """
        """
        return self.conn.execute('SELECT id FROM image WHERE path = ?', (image, )).fetchone() is not None


## TEST CODE ##

if __name__ == '__main__':
    model = DataModel({}, '/tmp/tagit.db')
    print model.has_changes()
    r = model.query(['hello', 'world'])
    print r

## EOF ##
