"""

> Input: model, candidates
* SVM ranks candidates
  > Input: candidates, features
  * Needs training
    > Input: candidates, features, true rank
    > Output: trained model
  * Features for each candidate
    > Input: model, candidates, candidate
    > Output: float
  > Output: candidate/rank
> Output: candidate/rank

Suggest new:
    * Model, images
    * Candidates from model, images
    * Apply some internal logic
    * Rank candidates
      * Compute features for each candidate wrt. model and other cands
      * Rank features with SVM

Suggest rank:
    * Model, candidates, history
    * Apply some internal logic
    * Rank candidates
      * Compute features for each candidate wrt. model, other cands, history
      * Rank features with SVM

Application:
    * selects features
    * Trained on data set
    * Compute ranks with SVM
    > Different params for different apps

Training:
    > Assume purely tag-based search (no metadata or other filters!)
    * Given training data
      One sample looks like this
        > Model
        > Candidates
        > Filters
        > Rank of each candidate
      Training data is a list of samples
        > Each sample has its own qid
        > Features are computed per candidate
    * Suggest new
        > Model:        Assignments at point t
        > Candidates:   All tags. If filters are given, minus all tags present in all resulting images
        > Filters:      From search log?
        > Ranks:        Tags according to timeline: Tag added after t -> Rank 1; Tag added after that -> Rank 2; ...
    * Suggest rank
        For each filter sequence (length M) in the search log,
        we can simulate the environment at every point N in that sequence.
        > Model:        Assignments at point t
        > Candidates:   Tags excluding filters
        > Filters:      N filters from search log
        > Ranks:        M-N filters from search log, from oldest to latest
    > When producing training data, check that they are balanced wrt. search history:
      There's many more samples with one filter tag than with two filter tags.
      Ensure uniform number of samples over filter length

Evaluation:
    * Given a trained model
    * Given a set of candidates
    > Compute features of candidates
    > Rank all candidates
    > Score: Return rank of candidate

Part of the tagit module.
A copy of the license is provided with the project.
Author: Matthias Baumgartner, 2017
"""
# imports
import svmlight
from os.path import exists as pexists
import warnings
import numpy as np

# inner-module imports
import features
from tagit.basics import unique
from tagit.external.memoize import memoized

# exports
__all__ = ('Score', )


## code ##

class Score(object):

    RANK_TARGET = 2 # Higher is better
    RANK_TAG    = 1

    def __init__(self, model_path, feature_def, model, candidates, filters=None):
        self.feature_def = feature_def
        self.candidates = candidates
        self.norm = (0.0, 1.0)

        # fetch class name from features, import them
        classes = unique([cls for cls, mth in map(lambda s: s.split('.'), feature_def)])

        self.instances = {}
        for classname in classes:
            #__import__('features', fromlist=[classname])
            cls = getattr(features, classname)
            inst = cls(model, candidates, filters)
            self.instances[classname] = inst

        if pexists(model_path):
            self.svmodel = svmlight.read_model(model_path)
        else:
            warnings.warn('No model loaded. Train first!')
            from os.path import join, dirname
            self.svmodel = svmlight.read_model(join(dirname(__file__), 'default_model.dat'))

    def _gen_features(self):
        funcs = []
        for feat in self.feature_def:
            classname, methodname = feat.split('.')
            fu = getattr(self.instances[classname], methodname)
            funcs.append(fu)
        return funcs

    def train(self, data):
        """
        data: {target -> [tags, ...]}
            target: tag which was selected (part of candidates!)
            tags:   list of candidates
            ...:    Additional info such as model, timestamp, history, ...
        qid:  iterate through targets
        rank: target has rank 2, others rank 1
        training data: [(enum, feature(tag))] per tag
        returns: function that stores model given path
        """
        # Get training data
        training_data = self._svm_features(data)
        # Learn normalization
        features = [map(snd, feat) for lbl, feat, qid in training_data]
        features = np.array(features)
        lo = features.min(axis=0)
        den = features.max(axis=0) - features.min(axis=0)
        self.norm = (lo, den)
        # Apply normalization
        features = ((features - lo) / den).tolist()
        training_data = [(lbl, list(enumerate(feat, 1)), qid) for feat, (lbl, fold, qid) in zip(features, training_data)]

        # Train SVM
        options = {'kernel': 'linear', 'verbosity': 0, 'type': 'ranking', 'C': 20.0}
        self.model = svmlight.learn(training_data, **options)
        return lambda path: svmlight.write_model(self.model, path) # TODO: path error checking due to svmlight's abort on failure

    def _svm_features(self, data):
        feat_fu = self._gen_features()
        training_data, qid = [], 0
        for target, tags in data.iteritems():
            qid += 1
            for t in tags:
                features = [f(t) for f in feat_fu] # Features of tag, enumeration for svmlight
                # normalize features
                lo, den = self.norm
                features = ((np.array(features) - lo) / den).tolist()
                features = list(enumerate(features, 1))
                lbl = t == target and self.RANK_TARGET or self.RANK_TAG
                training_data.append((lbl, features, qid))

        return training_data

    def prepare(self):
        # data: {target -> [candidates]}
        # classification mode -> rank/target don't matter
        # one set of candidates -> one target
        data = {None: self.candidates}
        test_data = self._svm_features(data)
        scores = svmlight.classify(self.svmodel, test_data)
        self.predictions = dict(zip(self.candidates, scores))
        return self

    @memoized
    def score(self, tag):
        if self.predictions is None: self.prepare()

        # pick score, return
        return self.predictions[tag]
        return sorted(self.candidates).index(tag)
        #return len(tag)
        #return max(map(len, self.candidates)) - len(tag)

## EOF ##
